class PSResult:
    """ Simple class holding results from power spectrum/2D FFT computation"""

    def __init__(self):
        self.spectra = None
        self.spec_name_list = None
        self.lb = None
        self.ps = None
        self.cov = None

    def store(self, spectra, spec_name_list, lb, ps, cov):
        self.spectra = spectra
        self.spec_name_list = spec_name_list
        self.lb = lb
        self.ps = ps
        self.cov = cov


class PSConfig:
    """ Simple class holding the configuration of the computation"""

    def __init__(self):
        self.lmax = None
        self.method = None
        self.error_method = None
        self.bin_size = None
        self.compute_T_only = None



from copy import deepcopy

import numpy as np
from scipy.ndimage.morphology import distance_transform_edt

from pixell import enmap
from pspy import (flat_tools, pspy_utils, so_cov, so_map, so_mcm, so_spectra,
                  so_window, sph_tools)


def create_window(
    patch,
    maps_info_list,
    apo_radius_survey,
    res_arcmin=0.5,
    galactic_mask=None,
    source_mask=None,
    compute_T_only=False,
    use_rmax=True,
):
    """Create a window function for a patch

    Parameters

    ----------
    patch: dict
      a dict containing the patch type and coordinates
      if patch_type is "Rectangle" the coordinates are expected to be the 4 corners
      if patch_type is "Disk" we expect the coordinates of the center and the radius in degree
    maps_info_list: list of dicts describing the data maps
      dictionnary should contain the name, the data type ("IQU" or "I") and optionally a calibration
      factor to apply to the map. Note that all map in the list should have the same data type
    apo_radius_survey: float
      the apodisation radius in degree (default: 1 degree)
    res_arcmin: float
      the angular resolution of the map in arcminutes (default: 0.5 arcminutes)
    source_mask: dict
      a dict containing an optional source mask and its properties
      the dictionnary should contain the name, the type of apodisation and the radius of apodisation
    galactic_mask: fits file
      an optional galactic mask to apply
    compute_T_only: boolean
      only use temperature field
    use_rmax: boolean
      apply apodization up to the apodization radius
    """

    if patch["patch_type"] == "Rectangle":
        car_box = patch["patch_coordinate"]
        window = so_map.read_map(maps_info_list[0]["name"], car_box=car_box)
        if maps_info_list[0]["data_type"] == "IQU":
            window.data = window.data[0]
            window.ncomp = 1
        window.data[:] = 0
        window.data[1:-1, 1:-1] = 1
        apo_type_survey = "C1"

    elif patch["patch_type"] == "Disk":
        dec_c, ra_c = patch["center"]
        radius = patch["radius"]
        eps = 0.1
        car_box = [
            [dec_c - radius - eps, ra_c - radius - eps],
            [dec_c + radius + eps, ra_c + radius + eps],
        ]
        window = so_map.read_map(maps_info_list[0]["name"], car_box=car_box)
        if maps_info_list[0]["data_type"] == "IQU":
            window.data = window.data[0]
            window.ncomp = 1

        window.data[:] = 1
        y_c, x_c = enmap.sky2pix(
            window.data.shape, window.data.wcs, [dec_c * np.pi / 180, ra_c * np.pi / 180]
        )
        window.data[int(y_c), int(x_c)] = 0
        dist = distance_transform_edt(window.data) * res_arcmin * 1 / 60
        window.data[dist < radius] = 0
        window.data = 1 - window.data
        apo_type_survey = "C1"

    if galactic_mask is not None:
        gal_mask = so_map.read_map(galactic_mask, car_box=car_box)
        window.data *= gal_mask.data
        del gal_mask

    for map_info in maps_info_list:
        split = so_map.read_map(map_info["name"], car_box=car_box)
        if compute_T_only and map_info["data_type"] == "IQU":
            split.data = split.data[0]
            split.ncomp = 1

        if split.ncomp == 1:
            window.data[split.data == 0] = 0.0

        else:
            for i in range(split.ncomp):
                window.data[split.data[i] == 0] = 0.0

    window = so_window.create_apodization(
        window, apo_type=apo_type_survey, apo_radius_degree=apo_radius_survey, use_rmax=use_rmax
    )

    if source_mask is not None:
        ps_mask = so_map.read_map(source_mask["name"], car_box=car_box)
        ps_mask = so_window.create_apodization(
            ps_mask, apo_type=source_mask["apo_type"], apo_radius_degree=source_mask["apo_radius"]
        )
        window.data *= ps_mask.data
        del ps_mask

    return car_box, window


def compute_mode_coupling(
    window,
    type,
    lmax,
    binning_file,
    ps_method="master",
    beam=None,
    lmax_pad=None,
    l_exact=None,
    l_band=None,
    l_toep=None,
    compute_T_only=False,
):
    """Compute the mode coupling corresponding the the window function

    Parameters
    ----------

    window: so_map
        the window function of the patch
    type: string
        the type of binning, either bin Cl or bin Dl
    lmax : integer
        the maximum multipole to consider for the spectra computation
    binning_file: text file
      a binning file with three columns bin low, bin high, bin mean
      note that either binning_file or bin_size should be provided
    ps_method: string
        the method for the computation of the power spectrum
        can be "master", "pseudo", or "2dflat" for now
    beam: text file
        file describing the beam of the map, expect bl to be the second column and start at l=0 (standard is : l,bl, ...)
    lmax_pad: integer
        the maximum multipole to consider for the mcm computation (optional)
        lmax_pad should always be greater than lmax
    save_coupling: str
    compute_T_only: boolean
        True to compute only T spectra
    """

    if ps_method == "2dflat":
        return None

    bin_lo, bin_hi, bin_c, bin_size = pspy_utils.read_binning_file(binning_file, lmax)
    n_bins = len(bin_hi)

    fsky = enmap.area(window.data.shape, window.data.wcs) / 4.0 / np.pi
    fsky *= np.mean(window.data)

    if beam is not None:
        beam_data = np.loadtxt(beam)
        if compute_T_only:
            beam = beam_data[:, 1]
        else:
            beam = (beam_data[:, 1], beam_data[:, 1])

    print("compute {} MCM".format(ps_method))
    if compute_T_only:
        if ps_method == "master":
            mbb_inv, Bbl = so_mcm.mcm_and_bbl_spin0(
                window,
                binning_file,
                bl1=beam,
                lmax=lmax,
                type=type,
                niter=0,
                lmax_pad=lmax_pad,
                l_exact=l_exact,
                l_band=l_band,
                l_toep=l_toep,
            )

        elif ps_method == "pseudo":
            mbb_inv = np.identity(n_bins)
            mbb_inv *= 1 / fsky
    else:
        window = (window, window)
        if ps_method == "master":
            mbb_inv, Bbl = so_mcm.mcm_and_bbl_spin0and2(
                window,
                binning_file,
                bl1=beam,
                lmax=lmax,
                type=type,
                niter=0,
                lmax_pad=lmax_pad,
                l_exact=l_exact,
                l_band=l_band,
                l_toep=l_toep,
            )

        elif ps_method == "pseudo":
            mbb_inv = {}
            spin_list = ["spin0xspin0", "spin0xspin2", "spin2xspin0"]
            for spin in spin_list:
                mbb_inv[spin] = np.identity(n_bins)
                mbb_inv[spin] *= 1 / fsky
            mbb_inv["spin2xspin2"] = np.identity(4 * n_bins)
            mbb_inv["spin2xspin2"] *= 1 / fsky

    return mbb_inv


def get_spectra(
    window,
    maps_info_list,
    car_box,
    type,
    lmax,
    binning_file,
    ps_method="master",
    mbb_inv=None,
    compute_T_only=False,
):
    """Compute the power spectra in the patch

    Parameters
    ----------
    window: so_map
        the window function of the patch
    maps_info_list: list of dicts describing the data maps
      dictionnary should contain the name, the data type ("IQU" or "I") and optionally a calibration factor to apply to the map
      note that all map in the list should have the same data type
    car_box: 2x2 array
      an array of the form [[dec0,rac0],[dec1,ra1]] it encompasses the patch
      and we will only load in memory the map inside the box
    type: string
        the type of binning, either bin Cl or bin Dl
    lmax : integer
        the maximum multipole to consider for the spectra computation
    ps_method: string
      the method for the computation of the power spectrum
      can be "master", "pseudo", or "2dflat" for now
    binning_file: text file
      a binning file with three columns bin low, bin high, bin mean
      note that either binning_file or bin_size should be provided
    mbb_inv: 2d array
      the inverse mode coupling matrix, not in use for 2dflat
    compute_T_only: boolean
        True to compute only T spectra

    """

    ht_list = []
    name_list = []

    if not compute_T_only:
        window = (window, window)

    for map_info in maps_info_list:

        split = so_map.read_map(map_info["name"], car_box=car_box)

        if compute_T_only and map_info["data_type"] == "IQU":
            split.data = split.data[0]
            split.ncomp = 1

        if map_info["cal"] is not None:
            split.data *= map_info["cal"]

        if ps_method in ["master", "pseudo"]:
            print("SPHT of %s in the patch" % map_info["name"])
            alms = sph_tools.get_alms(split, window, niter=0, lmax=lmax + 50)
            ht_list += [alms]

        elif ps_method == "2dflat":
            print("FFT of %s in the patch" % map_info["name"])
            ffts = flat_tools.get_ffts(split, window, lmax)
            ht_list += [ffts]

        name_list += [map_info["id"]]

    split_num = np.arange(len(maps_info_list))

    if compute_T_only:
        if ps_method in ["master", "pseudo"]:
            spectra = None
        elif ps_method == "2dflat":
            spectra = ["II"]
    else:
        if ps_method in ["master", "pseudo"]:
            spectra = ["TT", "TE", "TB", "ET", "BT", "EE", "EB", "BE", "BB"]
        elif ps_method == "2dflat":
            spectra = ["II", "IQ", "IU", "QI", "QQ", "QU", "UI", "UQ", "UU"]

    ps_dict = {}
    spec_name_list = []

    for name1, ht1, c1 in zip(name_list, ht_list, split_num):
        for name2, ht2, c2 in zip(name_list, ht_list, split_num):
            if c1 > c2:
                continue

            spec_name = "%sx%s" % (name1, name2)

            if ps_method in ["master", "pseudo"]:
                l, ps = so_spectra.get_spectra(ht1, ht2, spectra=spectra)
                ells, ps_dict[spec_name] = so_spectra.bin_spectra(
                    l, ps, binning_file, lmax, type=type, mbb_inv=mbb_inv, spectra=spectra
                )

            elif ps_method == "2dflat":
                ells, ps_dict[spec_name] = flat_tools.power_from_fft(ht1, ht2, type=type)

            spec_name_list += [spec_name]

    if compute_T_only:
        # to make TT only behave the same as the other cases, make it a dictionnary
        if ps_method in ["master", "pseudo"]:
            spectra = ["TT"]
            for spec_name in spec_name_list:
                ps_dict[spec_name] = {"TT": ps_dict[spec_name]}

    return spectra, spec_name_list, ells, ps_dict


def get_covariance(
    window,
    lmax,
    spec_name_list,
    ps_dict,
    binning_file,
    error_method="master",
    spectra=None,
    l_exact=None,
    l_band=None,
    l_toep=None,
    mbb_inv=None,
    compute_T_only=False,
):
    """Compute the covariance matrix of the power spectrum in the patch

    Parameters
    ----------

    window: so_map
      the window function of the patch
    lmax: integer
      the maximum multipole to consider for the spectra computation
    spec_name_list:  list
      the list of  power spectra
      For example : [split0xsplit0,split0xsplit1,split1xsplit1]
      note that for computing the error on PS(split0xsplit1) we need PS(split0xsplit0), PS(split0xsplit1), PS(split1xsplit1)
    ps_dict: dict
      a dict containing all power spectra
    binning_file: text file
      a binning file with three columns bin low, bin high, bin mean
      note that either binning_file or bin_size should be provided
    error_method: string
      the method for the computation of error
      can be "master" or "knox" for now
   approx_coupling: dict
   mbb_inv: 2d array
     the inverse mode coupling matrix, not in use for 2dflat
   compute_T_only: boolean
     True to compute only T spectra

    """

    bin_lo, bin_hi, bin_c, bin_size = pspy_utils.read_binning_file(binning_file, lmax)

    fsky = enmap.area(window.data.shape, window.data.wcs) / 4.0 / np.pi
    fsky *= np.mean(window.data)

    cov_dict = {}

    print("compute {} error".format(error_method))
    if error_method == "knox":
        for name in spec_name_list:
            m1, m2 = name.split("x")
            cov_dict[name] = {}
            for spec in spectra:
                X, Y = spec
                prefac = 1 / ((2 * bin_c + 1) * fsky * bin_size)
                # fmt: off
                cov_dict[name][X + Y] = np.diag(prefac * (
                    ps_dict["%sx%s" % (m1, m1)][X + X] * ps_dict["%sx%s" % (m2, m2)][Y + Y]
                    + ps_dict["%sx%s" % (m1, m2)][X + Y] ** 2
                ))
                # fmt: on

    elif error_method == "master":

        if not compute_T_only:
            mbb_inv = mbb_inv["spin0xspin0"]

        coupling_dict = so_cov.cov_coupling_spin0(
            window, lmax, niter=0, l_band=l_band, l_toep=l_toep, l_exact=l_exact
        )
        coupling = so_cov.bin_mat(coupling_dict["TaTcTbTd"], binning_file, lmax)

        for name in spec_name_list:
            m1, m2 = name.split("x")
            cov_dict[name] = {}
            for spec in spectra:
                X, Y = spec
                cov_dict[name][X + Y] = so_cov.symmetrize(
                    ps_dict["%sx%s" % (m1, m1)][X + X]
                ) * so_cov.symmetrize(ps_dict["%sx%s" % (m2, m2)][Y + Y])
                cov_dict[name][X + Y] += so_cov.symmetrize(ps_dict["%sx%s" % (m1, m2)][X + Y] ** 2)
                cov_dict[name][X + Y] *= coupling
                cov_dict[name][X + Y] = np.dot(np.dot(mbb_inv, cov_dict[name][X + Y]), mbb_inv.T)

    else:
        cov_dict = None

    return cov_dict


def theory_for_covariance(
    ps_dict, spec_name_list, spectra, lmax, beam=None, binning_file=None, force_positive=True
):

    ps_dict_for_cov = deepcopy(ps_dict)
    if force_positive:
        for name in spec_name_list:
            m1, m2 = name.split("x")
            for spec in spectra:
                X, Y = spec
                ps_dict_for_cov["%sx%s" % (m1, m1)][X + X] = np.abs(
                    ps_dict_for_cov["%sx%s" % (m1, m1)][X + X]
                )
                ps_dict_for_cov["%sx%s" % (m1, m1)][Y + Y] = np.abs(
                    ps_dict_for_cov["%sx%s" % (m2, m2)][Y + Y]
                )

    if beam is not None:
        beam_data = np.loadtxt(beam)
        l, bl = beam_data[:, 0], beam_data[:, 1]
        lb, bb = pspy_utils.naive_binning(l, bl, binning_file, lmax)
        for name in spec_name_list:
            for spec in spectra:
                ps_dict_for_cov[name][spec] *= bb ** 2

    return ps_dict_for_cov


def compute_ps(
    patch,
    maps_info_list,
    ps_method="master",
    error_method="master",
    type="Dl",
    binning_file=None,
    bin_size=None,
    beam=None,
    galactic_mask=None,
    source_mask=None,
    apo_radius_survey=1,
    compute_T_only=False,
    lmax=1000,
    lmax_pad=None,
    l_exact=None,
    l_band=None,
    l_toep=None,
):
    """Compute spectra

    Parameters
    ----------
    patch: dict
      a dict containing the patch type and coordinates
      if patch_type is "Rectangle" the coordinate are expected to be the 4 corners
      if patch_type is "Disk" we expect the coordinate of the center and the radius in degree
    maps_info_list: list of dicts describing the data maps
      dictionnary should contain the name, the data type ("IQU" or "I") and optionally a calibration factor to apply to the map
      note that all map in the list should have the same data type
    beam: text file
      file describing the beam of the map, expect l,bl
    binning_file: text file
      a binning file with three columns bin low, bin high, bin mean
      note that either binning_file or bin_size should be provided
    bin_size: integer
      the bin size
      note that either binning_file or bin_size should be provided
    type: string
      the type of binning, either bin Cl or bin Dl
    source_mask: dict
      a dict containing an optional source mask and its properties
      the dictionnary should contain the name, the type of apodisation and the radius of apodisation
    galactic_mask: fits file
      an optional galactic mask to apply
    apo_radius_survey: float
      the apodisation radius in degree
    ps_method: string
      the method for the computation of the power spectrum
      can be "master", "pseudo", or "2dflat" for now
    error_method: string
      the method for the computation of error
      can be "master" or "knox" for now
    compute_T_only: boolean
      True to compute only T spectra, should always be true for data_type= "I"
    lmax : integer
      the maximum multipole to consider for the spectra computation
    lmax_pad: integer
      the maximum multipole to consider for the mcm computation (optional)
      lmax_pad should always be greater than lmax
    """

    if binning_file is None:
        pspy_utils.create_binning_file(bin_size=bin_size, n_bins=1000, file_name="binning.dat")
        binning_file = "binning.dat"

    car_box, window = create_window(
        patch,
        maps_info_list,
        apo_radius_survey,
        galactic_mask=galactic_mask,
        source_mask=source_mask,
        compute_T_only=compute_T_only,
    )

    mbb_inv = compute_mode_coupling(
        window,
        type,
        lmax,
        binning_file,
        ps_method=ps_method,
        beam=beam,
        lmax_pad=lmax_pad,
        l_exact=l_exact,
        l_band=l_band,
        l_toep=l_toep,
        compute_T_only=compute_T_only,
    )

    spectra, spec_name_list, ells, ps_dict = get_spectra(
        window,
        maps_info_list,
        car_box,
        type,
        lmax,
        binning_file,
        ps_method=ps_method,
        mbb_inv=mbb_inv,
        compute_T_only=compute_T_only,
    )

    if ps_method == "2dflat" or error_method is None:
        return spectra, spec_name_list, ells, ps_dict, None

    ps_dict_for_cov = theory_for_covariance(
        ps_dict, spec_name_list, spectra, lmax, beam=beam, binning_file=binning_file
    )

    cov_dict = get_covariance(
        window,
        lmax,
        spec_name_list,
        ps_dict_for_cov,
        binning_file,
        error_method=error_method,
        l_exact=l_exact,
        l_band=l_band,
        l_toep=l_toep,
        spectra=spectra,
        mbb_inv=mbb_inv,
        compute_T_only=compute_T_only,
    )

    return spectra, spec_name_list, ells, ps_dict, cov_dict



# Copyright (c) Simons Observatory.
# Distributed under the terms of the Modified BSD License.
#
import os
from itertools import product

import ipywidgets as widgets
import numpy as np
import yaml

import plotly.graph_objects as go
from ipyleaflet import (DrawControl, FullScreenControl, LayersControl, Map,
                        MapStyle, WidgetControl)

from .ipyleaflet import (ColorizableTileLayer, Graticule, StatusBarControl,
                         allowed_colormaps)
from .ps_tools import compute_ps

# Default SO attribution for tiles
so_attribution = '&copy; <a href="https://simonsobservatory.org/">Simons Observatory</a>'


# Generate default plotly colormap based on planck colormap
def generate_default_colorscale(default="planck"):
    from pixell import colorize

    cs = colorize.schemes[default]
    return [[val, "rgb({},{},{})".format(*cols)] for val, cols in zip(cs.vals, cs.cols)]


default_colorscale = generate_default_colorscale()

# Output widget to catch functions/programs output into a widget
out = widgets.Output()


class App:
    """ An ipywidgets and plotly application for CMB map and power spectra visualization"""

    def __init__(self, config):
        if isinstance(config, dict):
            self.config = config
        else:
            with open(config, "r") as stream:
                self.config = yaml.load(stream, Loader=yaml.FullLoader)

        self.m = None
        self.p = None
        self.patches = dict()

        self.layers = [Graticule()]
        self._add_layers()
        self._add_map()
        self._add_plot()
        self._add_theory()

    def show_map(self):
        return self.m

    def show_plot(self):
        return self.p

    def _add_layers(self):
        def _get_section(section, name):
            value = section.get(name)
            if not value:
                raise ValueError("Missing '{}' section".format(name))
            return value

        self.maps_info_list = list()
        self.map_ids = list()
        maps = _get_section(self.config, "maps")
        for imap in maps:
            # Data info
            map_id = _get_section(imap, "id")
            self.map_ids.append(map_id)
            fits = _get_section(imap, "fits")
            data_type = imap.get("data_type", "IQU")
            info = {"name": fits, "data_type": data_type, "id": map_id, "cal": None}
            self.maps_info_list.append(info)

            # Tiles
            tiles = _get_section(imap, "tiles")
            path = _get_section(tiles, "path")

            # Set color range
            vrange = [[-500, +500], [-100, +100], [-100, +100]]
            _range = tiles.get("range")
            if _range:
                _val = _range.get("temperature")
                vrange[0] = [-_val, +_val] if _val else vrange[0]
                _val = _range.get("polarization")
                vrange[1] = [-_val, +_val] if _val else vrange[1]
                vrange[2] = vrange[1] if _val else vrange[2]
            for i, j in enumerate(["min", "max"]):
                _m = tiles.get(j)
                if _m:
                    _val = _m.get("temperature")
                    vrange[0][i] = _val if _val else vrange[0][i]
                    _val = _m.get("polarization")
                    vrange[1][i] = _val if _val else vrange[1][i]
                    vrange[2][i] = vrange[1][i] if _val else vrange[2][i]

            if any(s in path for s in [".png", "http"]):
                self.layers.append(
                    ColorizableTileLayer(
                        url=path,
                        base=True,
                        min_zoom=-5,
                        max_zoom=+5,
                        min_native_zoom=-5,
                        max_native_zoom=0,
                        tile_size=675,
                        attribution=tiles.get("attribution", so_attribution),
                        name=tiles.get("name", ""),
                        show_loading=False,
                        colormap=tiles.get("colormap", "planck"),
                        value_min=vrange[i][0],
                        value_max=vrange[i][1],
                    )
                )
            else:
                for i, item in enumerate(data_type):
                    name = "{} - {} - {}".format(tiles.get("prefix", "CMB"), map_id, item)
                    url = os.path.join("files", path, fits, "{z}/tile_{y}_{x}_%s.png" % i)
                    self.layers.append(
                        ColorizableTileLayer(
                            url=url,
                            base=True,
                            min_zoom=-5,
                            max_zoom=+5,
                            min_native_zoom=-5,
                            max_native_zoom=0,
                            tile_size=675,
                            attribution=tiles.get("attribution", so_attribution),
                            name=name,
                            show_loading=False,
                            colormap=tiles.get("colormap", "planck"),
                            value_min=vrange[i][0],
                            value_max=vrange[i][1],
                        )
                    )
        # if maps.get("sort", True):
        self.layers.sort()

    def _add_map(self):
        self.m = Map(
            layers=self.layers,
            controls=(FullScreenControl(), StatusBarControl(), LayersControl(position="topright"),),
            crs="CAR",
            center=(0, 0),
            min_zoom=-5,
            max_zoom=+5,
            interpolation="nearest",
            zoom=0,
            scroll_wheel_zoom=True,
            fade_animation=False,
            world_copy_jump=True,
            style=MapStyle(cursor="default"),
            default_style=MapStyle(cursor="default"),
            dragging_style=MapStyle(cursor="default"),
        )
        self._add_controls()

    def _add_controls(self):
        self.draw_control = DrawControl()
        self.draw_control.polyline = {}
        self.draw_control.polygon = {}
        self.draw_control.circlemarker = {}
        self.draw_control.circle = {"repeatMode": True}
        self.draw_control.rectangle = {"repeatMode": True}
        self.draw_control.edit = False
        self.draw_control.remove = False

        patches = self.patches

        def handle_draw(self, action, geo_json):
            if action == "created":
                name = "patch #{}".format(len(patches.keys()))
                patches[name] = geo_json
            elif action == "deleted":
                for k, v in patches.items():
                    if v == geo_json:
                        patches[k] = None
            else:
                print("Action = ", action)

        self.draw_control.on_draw(handle_draw)
        self.m.add_control(self.draw_control)

        scale = widgets.FloatSlider(
            value=1.0,
            min=0,
            max=2.0,
            step=0.01,
            description="scale",
            continuous_update=True,
            orientation="vertical",
            readout_format=".2f",
            layout=widgets.Layout(width="100px"),
        )

        def on_scale_change(change):
            for layer in self.layers:
                layer.scale = change["new"]

        scale.observe(on_scale_change, names="value")
        self.m.add_control(WidgetControl(widget=scale, position="bottomright"))

        cmap = widgets.RadioButtons(
            options=allowed_colormaps, value="planck", layout=widgets.Layout(width="100px"),
        )

        def on_cmap_change(change):
            for layer in self.layers:
                layer.colormap = change["new"]

        cmap.observe(on_cmap_change, names="value")
        self.m.add_control(WidgetControl(widget=cmap, position="bottomright"))

    def _add_theory(self):
        plot_config = self.config.get("plot", dict())
        clth = {}
        lth, clth["TT"], clth["EE"], clth["BB"], clth["TE"] = np.loadtxt(
            plot_config.get("theory", "bode_almost_wmap5_lmax_1e4_lensedCls_startAt2.dat"),
            unpack=True,
        )
        clth["ET"] = clth["TE"]
        for spec in ["EB", "BE", "BT", "TB"]:
            clth[spec] = clth["TE"] * 0.0
        self.theory = {"lth": lth, "clth": clth}

    def _add_plot(self):
        plot_config = self.config.get("plot", dict())
        plotly_config = plot_config.get("plotly", dict())

        self.fig = go.FigureWidget(
            layout=go.Layout(height=600, template=plotly_config.get("template", "plotly_white"))
        )

        # Header
        allowed_spectra = ["TT", "TE", "TB", "ET", "BT", "EE", "EB", "BE", "BB"]
        self.spectra_menu = widgets.Dropdown(description="Spectra:", options=allowed_spectra)
        self.split_menu = widgets.Dropdown(
            description="Split:",
            value="{}x{}".format(*self.map_ids),
            options=["{}x{}".format(*p) for p in product(self.map_ids, repeat=2)],
        )
        self.patch_menu = widgets.Dropdown(description="Patch:", options=self.patches.keys())
        self.header = widgets.HBox([self.spectra_menu, self.split_menu])

        # Config
        layout = widgets.Layout(width="auto", height="auto")
        self.compute_T_only = widgets.Checkbox(value=False, description="Temp. only", layout=layout)
        self.plot_theory = widgets.Checkbox(value=True, description="Plot theory", layout=layout)
        self.plot_theory.observe(self._update_theory, names="value")
        self.ps_method = widgets.RadioButtons(
            options=["master", "pseudo", "2dflat"],
            value=plot_config.get("method", "master"),
            description="Method:",
            layout=layout,
        )

        def _update_method(_):
            if self.ps_method.value == "2dflat":
                self.plot_theory.disabled = True
                self.header.children = tuple(list(self.header.children) + [self.patch_menu])
            else:
                self.plot_theory.disabled = False
                if len(self.header.children) == 3:
                    self.header.children = self.header.children[:-1]

        self.ps_method.observe(_update_method, names="value")

        self.error_method = widgets.RadioButtons(
            options=["master", "knox"],
            value=plot_config.get("error", "master"),
            description="Error:",
            layout=layout,
        )
        self.lmax = widgets.IntSlider(
            value=plot_config.get("lmax", 1000),
            min=0,
            max=10000,
            step=100,
            description=r"$\ell_\mathrm{max}$",
        )
        self.bin_size = widgets.IntSlider(
            value=plot_config.get("bin size", 40), min=0, max=100, step=10, description="Bin size",
        )
        container = widgets.HBox(
            [
                widgets.VBox([self.compute_T_only, self.plot_theory]),
                self.ps_method,
                self.error_method,
                widgets.VBox([self.lmax, self.bin_size]),
            ]
        )
        accordion = widgets.Accordion(children=[container, out], selected_index=None)
        accordion.set_title(0, "Configuration")
        accordion.set_title(1, "Logs")

        # Footer
        self.compute_button = widgets.Button(description="Compute spectra", icon="check")
        self.clean_button = widgets.Button(description="Clean patches", icon="trash-alt")

        def _clean_patches(_):
            for k in list(self.patches.keys()):
                del self.patches[k]
            self.draw_control.clear()
            self.clean_button.description = "Clean patches ({})".format(len(self.patches))

        self.compute_button.on_click(self._compute_spectra)
        self.clean_button.on_click(_clean_patches)
        footer = widgets.HBox([self.clean_button, self.compute_button])

        self.p = widgets.VBox([self.header, self.fig, accordion, footer])

    @out.capture()
    def _compute_spectra(self, _):
        out.clear_output()
        spectra = None

        self.compute_button.description = "Running..."
        self.compute_button.icon = "gear"
        self.compute_button.disabled = True
        self.clean_button.disabled = True
        self.clean_button.description = "Clean patches ({})".format(len(self.patches))

        def parse_rectangle(coordinates):
            return [coordinates[0][0][::-1], coordinates[0][2][::-1]]

        for ipatch, (name, patch) in enumerate(self.patches.items()):
            print("Compute patch #{}".format(ipatch))
            config = {
                "ps_method": self.ps_method.value,
                "error_method": self.error_method.value,
                "bin_size": self.bin_size.value,
                "compute_T_only": self.compute_T_only.value,
                "lmax": self.lmax.value,
            }
            if patch.get("config") == config:
                print("Patch already processed under the same condition")
                continue
            patch.update({"config": config})

            patch_dict = None
            geometry = patch.get("geometry")
            if geometry.get("type") == "Polygon":
                patch_dict = {
                    "patch_type": "Rectangle",
                    "patch_coordinate": parse_rectangle(geometry.get("coordinates")),
                }
            elif geometry.get("type") == "Point":
                style = patch.get("properties").get("style")
                if style.get("radius"):
                    patch_dict = {
                        "patch_type": "Disk",
                        "center": geometry.get("coordinates")[::-1],
                        "radius": style.get("radius"),
                    }
            else:
                print("Shape '{} not supported".format(geometry))
                continue

            spectra, spec_name_list, lb, ps_dict, cov_dict = compute_ps(
                patch_dict,
                self.maps_info_list,
                ps_method=self.ps_method.value,
                error_method=self.error_method.value,
                bin_size=self.bin_size.value,
                compute_T_only=self.compute_T_only.value,
                lmax=self.lmax.value,
            )
            patch.update(
                {
                    "results": {
                        "spectra": spectra,
                        "spec_name_list": spec_name_list,
                        "lb": lb,
                        "ps": ps_dict,
                        "cov": cov_dict,
                    }
                }
            )
        if spectra is not None:
            self.spectra_menu.options = spectra
            self.spectra_menu.observe(self._update_plot, names="value")
            self.split_menu.options = spec_name_list
            self.split_menu.observe(self._update_plot, names="value")
            self.patch_menu.options = self.patches.keys()
            self.patch_menu.observe(self._update_plot, names="value")

            self._update_plot(None)

        self.compute_button.description = "Compute spectra"
        self.compute_button.icon = "check"
        self.compute_button.disabled = False
        self.clean_button.disabled = False

    def _update_theory(self, _):
        if self.plot_theory.value:
            self.fig.add_scatter(
                name="theory",
                x=self.theory.get("lth")[: self.lmax.value],
                y=self.theory.get("clth")[self.spectra_menu.value][: self.lmax.value],
                mode="lines",
                line=dict(color="gray"),
            )
        else:
            # Remove last trace
            self.fig.data = self.fig.data[:-1]

    def _update_plot(self, _):
        split_name = self.split_menu.value
        spec = self.spectra_menu.value
        ps_method = self.ps_method.value

        # Clean data & layout
        self.fig.data = []
        # self.fig.layout = {}
        if ps_method == "2dflat":
            from plotly.colors import qualitative

            patch_name = self.patch_menu.value
            results = self.patches[patch_name].get("results")
            if not results:
                return
            powermap = results.get("ps").get(split_name).powermap[spec]
            powermap = np.fft.fftshift(powermap.copy())
            shape = powermap.shape
            self.fig.update_layout(
                width=self.fig.layout.height * shape[1] / shape[0],
                title=dict(
                    text="{} - {}".format(split_name, patch_name),
                    font=dict(color=qualitative.Plotly[list(self.patches).index(patch_name)]),
                ),
                xaxis_title=r"$\ell_X$",
                yaxis_title=r"$\ell_Y$",
                coloraxis={"colorscale": default_colorscale},
            )
            self.fig.add_heatmap(z=powermap, coloraxis="coloraxis")

        else:
            self.fig.update_layout(
                title=split_name, xaxis_title=r"$\ell$", yaxis_title=r"$D_\ell^\mathrm{%s}$" % spec,
            )

            for name, patch in self.patches.items():
                results = patch.get("results")
                if not results:
                    continue
                x = results.get("lb")
                y = results.get("ps").get(split_name).get(spec)
                yerr = np.sqrt(np.diag(results.get("cov").get(split_name).get(spec)))
                self.fig.add_scatter(
                    name=name,
                    x=x,
                    y=y,
                    error_y={"type": "data", "array": yerr, "visible": True},
                    mode="markers",
                )
            self._update_theory(None)



import glob
import os
import re

import numpy as np

from pixell import bunch, enmap, utils

default_pathformat = "tile%(y)03d_%(x)03d.fits"


def leaftile(
    idir, odir, tsize=675, comm=None, verbose=False, lrange=[0, -6], monolithic=False, slice=None
):
    """Given a input directory containing a tiled dmap in standard
    ordering, outputs a leaflet-compatible hierarchy of tiles in
    odir with tile size tsize."""
    # First create our base tiles. These have opposite y ordering than
    # dmap tiles, and may be different-sized.
    otilename = "tile_%(y)d_%(x)d.fits"
    if not monolithic:
        itilename = idir + "/tile%(y)03d_%(x)03d.fits"
        itile1, itile2 = None, None
    else:
        itilename = idir
        itile1, itile2 = (0, 0), (1, 1)
    retile(
        itilename,
        "%s/%d/%s" % (odir, lrange[0], otilename),
        ocorner=(np.pi / 2, -np.pi),
        otilesize=(-tsize, tsize),
        comm=comm,
        verbose=verbose,
        itile1=itile1,
        itile2=itile2,
        slice=slice,
    )
    # Then loop over the smaller levels
    for level in range(lrange[0] - 1, lrange[1], -1):
        if comm:
            comm.barrier()
        combine_tiles(
            "%s/%d/%s" % (odir, level + 1, otilename),
            "%s/%d/%s" % (odir, level, otilename),
            tyflip=True,
            pad_to=tsize,
            comm=comm,
            verbose=verbose,
        )


def combine_tiles(
    ipathfmt,
    opathfmt,
    combine=2,
    downsample=2,
    itile1=(None, None),
    itile2=(None, None),
    tyflip=False,
    txflip=False,
    pad_to=None,
    comm=None,
    verbose=False,
):
    """Given a set of tiles on disk at locaiton ipathfmt % {"y":...,"x"...},
    combine them into larger tiles, downsample and write the result to
    opathfmt % {"y":...,"x":...}. x and y must be contiguous and start at 0.

    reftile[2] indicates the tile coordinates of the first valid input tile.
    This needs to be specified if not all tiles of the logical tiling are
    physically present.

    tyflip and txflip indicate if the tiles coordinate system is reversed
    relative to the pixel coordinates or not."
    """
    # Expand combine and downsample to 2d
    combine = np.zeros(2, int) + combine
    downsample = np.zeros(2, int) + downsample
    if pad_to is not None:
        pad_to = np.zeros(2, int) + pad_to
    # Handle optional mpi
    rank, size = (comm.rank, comm.size) if comm is not None else (0, 1)
    # Find the range of input tiles
    itile1, itile2 = find_tile_range(ipathfmt, itile1, itile2)
    # Read the first tile to get its size information
    ibase = enmap.read_map(ipathfmt % {"y": itile1[0], "x": itile1[1]}) * 0
    # Find the set of output tiles we need to consider
    otile1 = itile1 // combine
    otile2 = (itile2 - 1) // combine + 1
    # And loop over them
    oyx = [(oy, ox) for oy in range(otile1[0], otile2[0]) for ox in range(otile1[1], otile2[1])]
    for i in range(rank, len(oyx), size):
        oy, ox = oyx[i]
        # Read in all associated tiles into a list of lists
        rows = []
        for dy in range(combine[0]):
            iy = oy * combine[0] + dy
            if iy >= itile2[0]:
                continue
            cols = []
            for dx in range(combine[1]):
                ix = ox * combine[1] + dx
                if ix >= itile2[1]:
                    continue
                if iy < itile1[0] or ix < itile1[1]:
                    # The first tiles are missing on disk, but are
                    # logically a part of the tiling. Use ibase,
                    # which has been zeroed out.
                    cols.append(ibase)
                else:
                    itname = ipathfmt % {"y": iy, "x": ix}
                    cols.append(enmap.read_map(itname))
            if txflip:
                cols = cols[::-1]
            rows.append(cols)
        # Stack them next to each other into a big tile
        if tyflip:
            rows = rows[::-1]
        omap = enmap.tile_maps(rows)
        # Downgrade if necessary
        if np.any(downsample > 1):
            omap = enmap.downgrade(omap, downsample)
        if pad_to is not None:
            # Padding happens towards the end of the tiling,
            # which depends on the flip status
            padding = np.array([[0, 0], [pad_to[0] - omap.shape[-2], pad_to[1] - omap.shape[-1]]])
            if tyflip:
                padding[:, 0] = padding[::-1, 0]
            if txflip:
                padding[:, 1] = padding[::-1, 1]
            omap = enmap.pad(omap, padding)
        # And output
        otname = opathfmt % {"y": oy, "x": ox}
        utils.mkdir(os.path.dirname(otname))
        enmap.write_map(otname, omap)
        if verbose:
            print(otname)


def retile(
    ipathfmt,
    opathfmt,
    itile1=(None, None),
    itile2=(None, None),
    otileoff=(0, 0),
    otilenum=(None, None),
    ocorner=(-np.pi / 2, -np.pi),
    otilesize=(675, 675),
    comm=None,
    verbose=False,
    slice=None,
    wrap=True,
):
    """Given a set of tiles on disk with locations ipathfmt % {"y":...,"x":...},
    retile them into a new tiling and write the result to opathfmt % {"y":...,"x":...}.
    The new tiling will have tile size given by otilesize[2]. Negative size means the
    tiling will to down/left instead of up/right. The corner of the tiling will
    be at sky coordinates ocorner[2] in radians. The new tiling will be pixel-
    compatible with the input tiling - w.g. the wcs will only differ by crpix.

    The output tiling will logically cover the whole sky, but only output tiles
    that overlap with input tiles will actually be written. This can be modified
    by using otileoff[2] and otilenum[2]. otileoff gives the tile indices of the
    corner tile, while otilenum indicates the number of tiles to write."""
    # Set up mpi
    rank, size = (comm.rank, comm.size) if comm is not None else (0, 1)
    # Expand any scalars
    if otilesize is None:
        otilesize = (675, 675)
    otilesize = np.zeros(2, int) + otilesize
    otileoff = np.zeros(2, int) + otileoff
    # Find the range of input tiles
    itile1, itile2 = find_tile_range(ipathfmt, itile1, itile2)
    # To fill in the rest of the information we need to know more
    # about the input tiling, so read the first tile
    ibase = enmap.read_map(ipathfmt % {"y": itile1[0], "x": itile1[1]})
    if slice:
        ibase = eval("ibase" + slice)
    itilesize = ibase.shape[-2:]
    ixres = ibase.wcs.wcs.cdelt[0]
    nphi = utils.nint(360 / np.abs(ixres))
    ntile_wrap = nphi // otilesize[1]
    # Find the pixel position of our output corners according to the wcs.
    # This is the last place we need to do a coordinate transformation.
    # All the rest can be done in pure pixel logic.
    pixoff = np.round(ibase.sky2pix(ocorner)).astype(int)

    # Find the range of output tiles
    def pix2otile(pix, ioff, osize):
        return (pix - ioff) // osize

    otile1 = pix2otile(itile1 * itilesize, pixoff, otilesize)
    otile2 = pix2otile(itile2 * itilesize - 1, pixoff, otilesize)
    otile1, otile2 = np.minimum(otile1, otile2), np.maximum(otile1, otile2)
    otile2 += 1
    # We can now loop over output tiles
    cache = [None, None, None]
    oyx = [(oy, ox) for oy in range(otile1[0], otile2[0]) for ox in range(otile1[1], otile2[1])]
    for i in range(rank, len(oyx), size):
        otile = np.array(oyx[i])
        # Find out which input tiles overlap with this output tile.
        # Our tile stretches from opix1:opix2 relative to the global input pixels
        opix1 = otile * otilesize + pixoff
        opix2 = (otile + 1) * otilesize + pixoff
        # output tiles and input tiles may increase in opposite directions
        opix1, opix2 = np.minimum(opix1, opix2), np.maximum(opix1, opix2)
        try:
            omap = read_area(
                ipathfmt, [opix1, opix2], itile1=itile1, itile2=itile2, cache=cache, slice=slice
            )
        except (IOError, OSError):
            continue
        x = otile[1] + otileoff[1]
        if wrap:
            x %= ntile_wrap
        oname = opathfmt % {"y": otile[0] + otileoff[0], "x": x}
        utils.mkdir(os.path.dirname(oname))
        enmap.write_map(oname, omap)
        if verbose:
            print(oname)


def read_monolithic(idir, verbose=True, slice=None, dtype=None):
    # Find the range of input tiles
    ipathfmt = idir + "/tile%(y)03d_%(x)03d.fits"
    itile1, itile2 = find_tile_range(ipathfmt)

    def read(fname):
        m = enmap.read_map(fname)
        if slice:
            m = eval("m" + slice)
        return m

    # Read the first and last tile to get the total dimensions
    m1 = read(ipathfmt % {"y": itile1[0], "x": itile1[1]})
    m2 = read(ipathfmt % {"y": itile2[0] - 1, "x": itile2[1] - 1})
    wy, wx = m1.shape[-2:]
    oshape = tuple(np.array(m1.shape[-2:]) * (itile2 - itile1 - 1) + np.array(m2.shape[-2:]))
    if dtype is None:
        dtype = m1.dtype
    omap = enmap.zeros(m1.shape[:-2] + oshape, m1.wcs, dtype)
    del m1, m2
    # Now loop through all tiles and copy them in to the correct position
    for ty in range(itile1[0], itile2[0]):
        for tx in range(itile1[1], itile2[1]):
            m = read(ipathfmt % {"y": ty, "x": tx})
            oy = ty - itile1[0]
            ox = tx - itile1[1]
            omap[..., oy * wy : (oy + 1) * wy, ox * wx : (ox + 1) * wx] = m  # noqa
            if verbose:
                print(ipathfmt % {"y": ty, "x": tx})
    return omap


def monolithic(idir, ofile, verbose=True, slice=None, dtype=None):
    omap = read_monolithic(idir, verbose=verbose, slice=slice, dtype=dtype)
    enmap.write_map(ofile, omap)


def range_overlap(a, b):
    return np.array([np.maximum(a[0], b[0]), np.minimum(a[1], b[1])])


def find_tile_range(pathfmt, tile1=(None, None), tile2=(None, None)):
    """Given a path format with with labeled formats including y
    and x (like %(y)d), Return the range of tiles available, in
    the form of [{start,end},{y,x}] tile indices.
    If tile1[2] is specified, then this will override the start
    of the result. If tile2[2] is specified, then it will override
    the end of the result."""
    if tile1 is None:
        tile1 = (None, None)
    if tile2 is None:
        tile2 = (None, None)
    # If both offset and ntile are specified, we don't need any
    # complicated disk search.
    if (
        tile1[0] is not None
        and tile1[1] is not None
        and tile2[0] is not None
        and tile2[1] is not None
    ):
        return np.asarray(tile1), np.asarray(tile2)
    # Find the min/max on disk. We do that by constructing a glob to
    # roughly match them, and them filtering them with a regex.
    ranges = [None, None]
    gstr = utils.format_to_glob(pathfmt)
    regex = utils.format_to_regex(pathfmt)
    files = glob.glob(gstr)
    if len(files) == 0:
        raise ValueError("Found no files matching path format!")
    for file in files:
        m = re.match(regex, file)
        if not m:
            continue
        try:
            yx = [int(m.group(name)) for name in ["y", "x"]]
        except IndexError:
            yx = [0, 0]
        for i in range(2):
            if ranges[i] is None:
                ranges[i] = [yx[i], yx[i] + 1]
            ranges[i] = [min(ranges[i][0], yx[i]), max(ranges[i][1], yx[i] + 1)]
    # Override if needed
    for i in range(2):
        if tile1[i] is not None:
            ranges[i][0] = tile1[i]
        if tile2[i] is not None:
            ranges[i][1] = tile2[i]
    ranges = np.array(ranges)
    return ranges[:, 0], ranges[:, 1]


def read_tileset_geometry(ipathfmt, itile1=(None, None), itile2=(None, None)):
    itile1, itile2 = find_tile_range(ipathfmt, itile1, itile2)
    mfile1 = ipathfmt % {"y": itile1[0], "x": itile1[1]}
    mfile2 = ipathfmt % {"y": itile2[0] - 1, "x": itile2[1] - 1}
    m1 = enmap.read_map(mfile1)
    m2 = m1 if mfile1 == mfile2 else enmap.read_map(mfile2)
    wy, wx = m1.shape[-2:]
    oshape = tuple(np.array(m1.shape[-2:]) * (itile2 - itile1 - 1) + np.array(m2.shape[-2:]))
    return bunch.Bunch(
        shape=m1.shape[:-2] + oshape, wcs=m1.wcs, dtype=m1.dtype, tshape=m1.shape[-2:]
    )


def read_area(
    ipathfmt,
    opix,
    itile1=(None, None),
    itile2=(None, None),
    verbose=False,
    cache=None,
    slice=None,
    wrap=True,
):
    """Given a set of tiles on disk with locations ipathfmt % {"y":...,"x":...},
    read the data corresponding to the pixel range opix[{from,to},{y,x}] in
    the full map."""
    opix = np.asarray(opix)
    # Find the range of input tiles
    itile1, itile2 = find_tile_range(ipathfmt, itile1, itile2)
    # To fill in the rest of the information we need to know more
    # about the input tiling, so read the first tile
    if cache is None or cache[2] is None:
        geo = read_tileset_geometry(ipathfmt, itile1=itile1, itile2=itile2)
    else:
        geo = cache[2]
    if cache is not None:
        cache[2] = geo
    # Determine tile wrapping
    npix_phi = np.abs(360.0 / geo.wcs.wcs.cdelt[0])
    ntile_phi = utils.nint(npix_phi / geo.tshape[-1])

    isize = geo.tshape
    osize = opix[1] - opix[0]
    omap = enmap.zeros(geo.shape[:-2] + tuple(osize), geo.wcs, geo.dtype)
    # Find out which input tiles overlap with this output tile.
    # Our tile stretches from opix1:opix2 relative to the global input pixels
    it1 = opix[0] // isize
    it2 = (opix[1] - 1) // isize + 1
    noverlap = 0
    for ity in range(it1[0], it2[0]):
        if ity < itile1[0] or ity >= itile2[0]:
            continue
        # Start/end of this tile in global input pixels
        ipy1, ipy2 = ity * isize[0], (ity + 1) * isize[0]
        overlap = range_overlap(opix[:, 0], [ipy1, ipy2])
        oy1, oy2 = overlap - opix[0, 0]
        iy1, iy2 = overlap - ipy1
        for itx in range(it1[1], it2[1]):
            if wrap:
                itx_wrap = itx % ntile_phi
            if itx_wrap < itile1[1] or itx_wrap >= itile2[1]:
                continue
            ipx1, ipx2 = itx * isize[1], (itx + 1) * isize[1]
            overlap = range_overlap(opix[:, 1], [ipx1, ipx2])
            ox1, ox2 = overlap - opix[0, 1]
            ix1, ix2 = overlap - ipx1
            # Read the input tile and copy over
            iname = ipathfmt % {"y": ity, "x": itx_wrap}
            if cache is None or cache[0] != iname:
                imap = enmap.read_map(iname)
                if slice:
                    imap = eval("imap" + slice)
            else:
                imap = cache[1]
            if cache is not None:
                cache[0], cache[1] = iname, imap
            if verbose:
                print(iname)
            # Edge input tiles may be smaller than the standard
            # size.
            ysub = isize[0] - imap.shape[-2]
            xsub = isize[1] - imap.shape[-1]
            # If the input map is too small, there may actually be
            # zero overlap.
            if oy2 - ysub <= oy1 or ox2 - xsub <= ox1:
                continue
            # fmt: off
            omap[..., oy1 : oy2 - ysub, ox1 : ox2 - xsub] = imap[..., iy1 : iy2 - ysub, ix1 : ix2 - xsub] # noqa
            # fmt: on
            noverlap += 1
    if noverlap == 0:
        raise IOError(
            "No tiles for tiling %s in range %s"
            % (ipathfmt, ",".join([":".join([str(p) for p in r]) for r in opix.T]))
        )
    # Set up the wcs for the output tile
    omap.wcs.wcs.crpix -= opix[0, ::-1]
    return omap


def read_retile(
    ipathfmt,
    tpos,
    otilesize=None,
    pixoff=(0, 0),
    margin=0,
    itile1=(None, None),
    itile2=(None, None),
    verbose=False,
):
    """Read a single tile from the tiling at ipathfmt % {"y":tpos[0],"x":tpos[1]},
    returning it as an enmap. If otilesize or pixoff are specified, then
    the tiling will be from a retiling with that tile size and pixel offset,
    which means that behind the scenes multiple tiles will be read and stitched
    together. If margin[{left,right},{y,x}] is not zero, then it specifies the
    number of pixels to extend the tile by. The tile will be extended with
    data from the tiling, i.e. from the neighboring tiles."""
    if otilesize is None:
        geo = read_tileset_geometry(ipathfmt, itile1, itile2)
        otilesize = geo.tshape
    tpos = np.zeros(2, int) + tpos
    pixoff = np.zeros(2, int) + pixoff
    margin = np.zeros((2, 2), int) + margin
    pixbox = np.array([tpos * otilesize, (tpos + 1) * otilesize]) + pixoff
    pixbox[0] -= margin[0]
    pixbox[1] += margin[1]
    return read_area(ipathfmt, pixbox, itile1=itile1, itile2=itile2, verbose=verbose)


def retile_iterator(
    ipathfmt,
    otilesize=None,
    pixoff=(0, 0),
    margin=0,
    itile1=(None, None),
    itile2=(None, None),
    comm=None,
    verbose=False,
):
    """Iterator that yields a series of tiles from the tileset given by
    ipathfmt. See read_retile for how margin, otilesize and pixoff
    allow you to iterate over a modified tiling. This currently just
    calls read_area repeatedly, so it is not as efficient as it could have
    been."""
    # Handle mpi
    rank, nproc = (0, 1) if comm is None else (comm.rank, comm.size)
    # Find the number of tiles to iterate over
    geo = read_tileset_geometry(ipathfmt, itile1, itile2)
    if otilesize is None:
        otilesize = geo.tshape
    otilesize = np.array(otilesize)
    notile = (np.array(geo.shape[-2:]) - pixoff + otilesize - 1) // otilesize
    tyx = [(y, x) for y in range(notile[0]) for x in range(notile[1])]
    for i in range(rank, len(tyx), nproc):
        tpos = tyx[i]
        yield tpos, read_retile(
            ipathfmt,
            tpos,
            otilesize=otilesize,
            pixoff=pixoff,
            margin=margin,
            itile1=itile1,
            itile2=itile2,
        )



import numpy as np

from pspy import so_map


def healpix2car(
    input_file,
    fields=(0, 1, 2),
    mask_file=None,
    output_file=None,
    resolution=0.5,
    bounding_box=(-180, 180, -75, 30),
    lmax=6000,
):
    """ Convert HEALPIX map to CAR map
    Parameters
    ----------
    input_file: fits file
      name of the input HEALPIX fits file
    fields: tuple
      HEALPIX fields to convert i.e. (0,) will keep only temperature field
    mask_file: fits file
      name of the HEALPIX mask file
    output_file: fits file
      name of the output CAR fits file
    resolution: string
      CAR final resolution in arcminutes
    bounding_box:  tuple
      (ra0, ra1, dec0, dec1) in degree

    """
    healpix_map = so_map.read_map(input_file, fields_healpix=fields)

    # CAR Template
    ra0, ra1, dec0, dec1 = bounding_box
    res = resolution
    car_template = so_map.car_template(healpix_map.ncomp, ra0, ra1, dec0, dec1, res)
    projected_map = so_map.healpix2car(healpix_map, car_template, lmax=lmax)

    if mask_file is not None:
        mask = so_map.read_map(mask_file)
        projected_mask = so_map.healpix2car(mask, car_template, lmax=lmax)
        if mask.ncomp == healpix_map.ncomp == 1:
            projected_map.data *= np.where(projected_mask.data < 0.5, 0, 1)
        elif mask.ncomp == 1:
            for i in range(healpix_map.ncomp):
                projected_map.data[i] *= np.where(projected_mask.data < 0.5, 0, 1)
        else:
            if healpix_map.ncomp != mask.ncomp:
                raise ValueError("Map and mask have different number of components")
            for i in range(mask.ncomp):
                projected_map.data[i] *= np.where(projected_mask.data[i] < 0.5, 0, 1)

    print("Writing '{}' file".format(output_file))
    projected_map.write_map(output_file)
    return projected_map


def main():
    import argparse

    parser = argparse.ArgumentParser(
        description="A python program to convert HEALPIX map into CAR map"
    )
    parser.add_argument(
        "-i",
        "--input-file",
        help="input FITS file corresponding to HEALPIX map",
        type=str,
        required=True,
        default=None,
    )
    parser.add_argument(
        "-o",
        "--output-file",
        help="output FITS file corresponding to CAR map",
        type=str,
        required=True,
        default=None,
    )
    parser.add_argument(
        "-f",
        "--fields",
        help="tuple that enables HEALPIX fields i.e. (0,) will only keep temperature field ",
        type=tuple,
        default=(0, 1, 2),
    )
    parser.add_argument(
        "--bounding-box",
        help="set the bounding box (ra0, dec0, ra1, dec1) all in degrees",
        type=tuple,
        default=(-180, 180, -75, 30),
    )
    parser.add_argument(
        "--resolution", help="set the resolution in arcminutes", type=float, default=0.5
    )
    parser.add_argument(
        "--mask-file",
        help="set a mask file to apply to HEALPIX map before converting",
        type=str,
        default=None,
    )
    args = parser.parse_args()

    healpix2car(
        input_file=args.input_file,
        fields=[int(i) for i in args.fields],
        mask_file=args.mask_file,
        output_file=args.output_file,
        resolution=args.resolution,
        bounding_box=args.bounding_box,
    )


# script:
if __name__ == "__main__":
    main()



# Copyright (c) Simons Observatory.
# Distributed under the terms of the Modified BSD License.
#
from traitlets import CFloat, Enum, Unicode

from ipyleaflet import Control, Layer, LocalTileLayer, allowed_crs

from ._version import EXTENSION_VERSION

allowed_crs += ['CAR']
allowed_colormaps = ['gray', 'planck', 'wmap', 'hotcold']


class Graticule(Layer):
    _view_name = Unicode('LeafletGraticuleView').tag(sync=True)
    _model_name = Unicode('LeafletGraticuleModel').tag(sync=True)
    _view_module = Unicode('jupyter-leaflet-car').tag(sync=True)
    _model_module = Unicode('jupyter-leaflet-car').tag(sync=True)

    _view_module_version = Unicode(EXTENSION_VERSION).tag(sync=True)
    _model_module_version = Unicode(EXTENSION_VERSION).tag(sync=True)

    name = Unicode('graticule').tag(sync=True)


class ColorizableTileLayer(LocalTileLayer):
    _view_name = Unicode('LeafletColorizableTileLayerView').tag(sync=True)
    _model_name = Unicode('LeafletColorizableTileLayerModel').tag(sync=True)
    _view_module = Unicode('jupyter-leaflet-car').tag(sync=True)
    _model_module = Unicode('jupyter-leaflet-car').tag(sync=True)

    _view_module_version = Unicode(EXTENSION_VERSION).tag(sync=True)
    _model_module_version = Unicode(EXTENSION_VERSION).tag(sync=True)

    colormap = Enum(values=allowed_colormaps, default_value='planck').tag(sync=True, o=True)
    value_min = CFloat(-500).tag(sync=True, o=True)
    value_max = CFloat(+500).tag(sync=True, o=True)
    scale = CFloat(1.0).tag(sync=True, o=True)

    def __lt__(self, other):
        return self.name[-1] < other.name[-1]


class StatusBarControl(Control):
    _view_name = Unicode('LeafletStatusBarControlView').tag(sync=True)
    _model_name = Unicode('LeafletStatusBarControlModel').tag(sync=True)
    _view_module = Unicode('jupyter-leaflet-car').tag(sync=True)
    _model_module = Unicode('jupyter-leaflet-car').tag(sync=True)

    _view_module_version = Unicode(EXTENSION_VERSION).tag(sync=True)
    _model_module_version = Unicode(EXTENSION_VERSION).tag(sync=True)

    prefix = Unicode('').tag(sync=True, o=True)
    position = Unicode('bottomleft').tag(sync=True, o=True)



import os

import numpy as np

from pixell import enplot, mpi, wcsutils
from pspy import so_map

from . import tile_utils_sigurd, webplot


def car2tiles(
    input_file,
    mask_file=None,
    enplot_args=None,
    output_dir=None,
    delete_fits=True,
    use_webplot=True,
):
    """ Convert CAR map to PNG tiles
    Parameters
    ----------
    input_file: fits file
      name of the input CAR fits file
    mask_file: fits file
      name of the CAR mask file
    enplot_args:
      list of enplot/webplot options (see corresponding programs)
    output_dir: string
      name of the output directory holding PNG files
    delete_fits: boolean
      delete the FITS files corresponding to the tiles
    use_webplot: boolean
      use webplot in place of enplot program
    """
    enplot_args = enplot_args or []

    comm = mpi.COMM_WORLD
    if comm.rank == 0:
        # Check if path to fits file are already stored
        fits_files = os.path.join(output_dir, "*/*.fits")
        if fits_files not in enplot_args:
            enplot_args.append(fits_files)

        if output_dir is None:
            output_dir = os.path.join("leaflet", input_file)

        if os.path.exists(output_dir):
            os.system("rm -rf %s" % output_dir)

        if mask_file is not None:
            car = so_map.read_map(input_file)
            mask = so_map.read_map(mask_file)

            if not wcsutils.is_compatible(mask.geometry[0], car.geometry[0]):
                raise ValueError("Map and mask must have compatible geometries")

            if mask.ncomp == car.ncomp == 1:
                car.data *= np.where(mask.data < 0.5, 0, 1)
            elif mask.ncomp == 1:
                for i in range(car.ncomp):
                    car.data[i] *= np.where(mask.data < 0.5, 0, 1)
            else:
                if mask.ncomp != car.ncomp:
                    raise ValueError("Map and mask have different number of components")
                for i in range(mask.ncomp):
                    car.data[i] *= np.where(mask.data[i] < 0.5, 0, 1)
            input_file += ".tmp"
            car.write_map(input_file)

    if not mpi.disabled:
        comm.barrier()

    tile_utils_sigurd.leaftile(
        input_file,
        output_dir,
        verbose="-v" in enplot_args,
        comm=comm if not mpi.disabled else None,
        monolithic=True,
    )

    if use_webplot:
        args = webplot.parse_args(enplot_args)
        webplot.plot(args)
    else:
        args = enplot.parse_args(enplot_args)
        for plot in enplot.plot_iterator(*args.ifiles, comm=comm, **args):
            enplot.write(plot.name, plot)

    if comm.rank == 0:
        if mask_file is not None:
            os.remove(input_file)

        if delete_fits:
            for fits in args.ifiles:
                os.remove(fits)


def main():
    import argparse

    parser = argparse.ArgumentParser(
        description="A python program to produce Sigurd's plots and corresponding html pages",
        epilog="Additional options will be passed to enplot/webplot program",
    )
    parser.add_argument(
        "-i",
        "--input-file",
        help="input FITS file corresponding to CAR map",
        type=str,
        required=True,
        default=None,
    )
    parser.add_argument(
        "--mask-file",
        help="set a mask file to apply to HEALPIX map before converting",
        type=str,
        default=None,
    )
    parser.add_argument(
        "--output-dir", help="output directory holding png files", type=str, default="leaflet"
    )
    parser.add_argument(
        "--use-enplot",
        help="use enplot routine (default use webplot)",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--keep-fits-files", help="keep intermediate FITS files", action="store_true", default=False
    )
    args, enplot_args = parser.parse_known_args()

    car2tiles(
        args.input_file,
        args.mask_file,
        enplot_args,
        output_dir=args.output_dir,
        delete_fits=not args.keep_fits_files,
        use_webplot=not args.use_enplot,
    )


# script:
if __name__ == "__main__":
    main()



from ._version import __version__, version_info  # noqa
from .psplay import App  # noqa


def _jupyter_nbextension_paths():
    return [
        {
            "section": "notebook",
            "src": "static",
            "dest": "jupyter-leaflet-car",
            "require": "jupyter-leaflet-car/extension",
        }
    ]



# -*- coding: utf-8 -*-

import json
import os
import sys
from distutils import log
from glob import glob
from subprocess import check_call

from setuptools import Command, find_packages, setup
from setuptools.command.build_py import build_py
from setuptools.command.egg_info import egg_info
from setuptools.command.sdist import sdist

here = os.path.dirname(os.path.abspath(__file__))
node_root = os.path.join(here, "js")
is_repo = os.path.exists(os.path.join(here, ".git"))

npm_path = os.pathsep.join(
    [os.path.join(node_root, "node_modules", ".bin"), os.environ.get("PATH", os.defpath),]
)

log.set_verbosity(log.DEBUG)
log.info("setup.py entered")
log.info("$PATH=%s" % os.environ["PATH"])


def js_prerelease(command, strict=False):
    """Decorator for building minified js/css prior to another command."""

    class DecoratedCommand(command):
        def run(self):
            jsdeps = self.distribution.get_command_obj("jsdeps")
            if not is_repo and all(os.path.exists(t) for t in jsdeps.targets):
                # sdist, nothing to do
                command.run(self)
                return

            try:
                self.distribution.run_command("jsdeps")
            except Exception as e:
                missing = [t for t in jsdeps.targets if not os.path.exists(t)]
                if strict or missing:
                    log.warn("rebuilding js and css failed")
                    if missing:
                        log.error("missing files: %s" % missing)
                    raise e
                else:
                    log.warn("rebuilding js and css failed (not a problem)")
                    log.warn(str(e))
            command.run(self)
            update_package_data(self.distribution)

    return DecoratedCommand


def get_data_files():
    with open(os.path.join(node_root, "package.json")) as f:
        package_json = json.load(f)
    tgz = "%s-%s.tgz" % (package_json["name"], package_json["version"])

    return [
        ("share/jupyter/nbextensions/jupyter-leaflet-car", glob("psplay/static/*")),
        ("etc/jupyter/nbconfig/notebook.d", ["jupyter-leaflet-car.json"]),
        ("share/jupyter/lab/extensions", ["js/" + tgz]),
    ]


def update_package_data(distribution):
    """Update package_data to catch changes during setup."""
    build_py = distribution.get_command_obj("build_py")
    distribution.data_files = get_data_files()
    # re-init build_py options which load package_data
    build_py.finalize_options()


class NPM(Command):
    description = "install package.json dependencies using npm"

    user_options = []

    node_modules = os.path.join(node_root, "node_modules")

    targets = [
        os.path.join(here, "psplay", "static", "extension.js"),
        os.path.join(here, "psplay", "static", "index.js"),
    ]

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def has_npm(self):
        try:
            check_call(["npm", "--version"])
            return True
        except:
            return False

    def run(self):
        has_npm = self.has_npm()
        if not has_npm:
            log.error(
                "`npm` unavailable.  If you're running this command using sudo, make sure `npm` is available to sudo"
            )

        env = os.environ.copy()
        env["PATH"] = npm_path

        if self.has_npm():
            log.info("Installing build dependencies with npm.  This may take a while...")
            check_call(["npm", "install"], cwd=node_root, stdout=sys.stdout, stderr=sys.stderr)
            check_call(["npm", "pack"], cwd=node_root, stdout=sys.stdout, stderr=sys.stderr)

        for t in self.targets:
            if not os.path.exists(t):
                msg = "Missing file: %s" % t
                if not has_npm:
                    msg += "\nnpm is required to build a development version of widgetsnbextension"
                raise ValueError(msg)

        # update package data in case this created new files
        update_package_data(self.distribution)


with open("README.rst") as readme_file:
    readme = readme_file.read()

setup_args = {
    "name": "psplay",
    "version": "0.0.3",
    "description": "A Jupyter widget for playing with CAR Leaflet maps",
    "license": "BSD License",
    "include_package_data": True,
    "data_files": get_data_files(),
    "install_requires": [
        "ipywidgets>=7.5.0",
        "traittypes>=0.2.1",
        "ipyleaflet==0.12.4",
        "pspy>=1.3.1",
        "plotly>=4.6.0",
        "pyyaml",
    ],
    "packages": find_packages(),
    "zip_safe": False,
    "cmdclass": {
        "build_py": js_prerelease(build_py),
        "egg_info": js_prerelease(egg_info),
        "sdist": js_prerelease(sdist, strict=True),
        "jsdeps": NPM,
    },
    "author": "Simons Observatory Collaboration Power Spectrum Task Force",
    "url": "https://github.com/simonsobs/psplay",
    "long_description": readme,
    "keywords": ["CMB", "spectra", "ipython", "jupyter", "widgets", "graphics"],
    "classifiers": [
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Multimedia :: Graphics",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    "entry_points": {
        "console_scripts": [
            "car2tiles=psplay.tools.car2tiles:main",
            "healpix2car=psplay.tools.healpix2car:main",
        ],
    },
}

setup(**setup_args)



import argparse
import glob
import shlex

import numpy as np
from PIL import Image

from pixell import bunch, enmap, mpi


def define_arg_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("ifiles", nargs="+")
    parser.add_argument("-b", "--nbyte", type=int, default=4)
    parser.add_argument(
        "-q",
        "--quantum",
        type=float,
        default=1,
        help="""The quantization to use in the fixed-point representation of the image data. Higher values reduce
      image size by discarding more noise and allows higher extremes in the map, but reduces the fidelity
      of the map. 1 uK, the current default, effectively increases the noise by 0.5% in the very deepest
      parts of our maps, and by less in shallower regions. Compared to a more conservative 0.1 it results
      in 25% smaller images. The quantization level is stored in the high bit of the first 32 pixels of
      the image.""",
    )
    parser.add_argument("--suffix", type=str, default="")
    parser.add_argument("--ext", type=str, default=".png")
    parser.add_argument("-v", "--verbose", action="count", default=0)
    parser.add_argument(
        "-m",
        "--mask",
        type=float,
        default=0,
        help="Treat values exactly equal to this floating point value as masked",
    )
    return parser


arg_parser = define_arg_parser()


def parse_args(args, noglob=False):
    if isinstance(args, str):
        args = shlex.split(args)
    res, unkown = arg_parser.parse_known_args(args)
    res = bunch.Bunch(**res.__dict__)
    # Glob expansion
    if not noglob:
        ifiles = []
        for pattern in res.ifiles:
            matches = glob.glob(pattern)
            if len(matches) > 0:
                ifiles += matches
            else:
                ifiles.append(pattern)
        res.ifiles = ifiles
    return res


def pack(imap, mask, nbyte=4, quantum=1.0):
    # Flip y to match PIL pixel origin in top left corner
    imap = imap[..., ::-1, :]
    mask = mask[::-1]
    # Quantize
    qmap = np.round(imap / quantum).astype(np.int64)
    # Switch from twos complement to lest significant sign and mag
    neg = qmap < 0
    qmap[neg] = -qmap[neg]
    qmap = qmap.view(np.uint64)
    qmap <<= 1
    qmap |= neg
    # Mark masked values as all ff
    qmap[mask] = 0xFFFFFFFFFFFFFFFF
    # Express as bytes and truncate to requested number of bytes
    qmap = qmap.view(np.uint8).reshape(imap.shape + (8,))[..., :nbyte]
    # Reformat as byte planes
    qmap = np.moveaxis(qmap, -1, -3)
    # Stack planes in the y direction
    qmap = qmap.reshape(qmap.shape[:-3] + (-1, qmap.shape[-1]))
    # Add metadata row
    meta = np.concatenate(
        [np.array([nbyte], np.uint8), np.array([quantum], np.float64).view(np.uint8)]
    )
    omap = np.zeros(qmap.shape[:-2] + (qmap.shape[-2] + 1, qmap.shape[-1]), np.uint8)
    omap[..., 1:, :] = qmap
    omap[..., 0, : len(meta)] = meta
    return omap


def unpack(imap):
    # Read the metadata row
    meta, qmap = imap[0], imap[1:]
    nbyte = meta[0]
    quantum = meta[1:5].view(np.float32)[0]
    # Undo plane stacking
    qmap = qmap.reshape(nbyte, -1, qmap.shape[1])
    qmap = qmap.moveaxis(0, -1)
    mask = np.all(qmap == 0xFF, -1)
    wmap = np.zeros(qmap.shape[:2], 8)
    wmap[:, :, :nbyte] = qmap
    wmap = wmap.view(np.uint64)
    # Back to twos complement
    neg = (wmap & 1) == 1
    wmap >>= 1
    wmap = wmap.view(np.int64)
    wmap[neg] = -wmap[neg]
    # And to real units
    omap = wmap * quantum
    omap[mask] = 0
    omap, mask = omap[::-1], mask[::-1]
    return omap, mask


def plot(args):
    def get_num_digits(n):
        return int(np.log10(n)) + 1

    comm = mpi.COMM_WORLD
    ifiles = sum([sorted(glob.glob(ifile)) for ifile in args.ifiles], [])

    for ind in range(comm.rank, len(ifiles), comm.size):
        ifile = ifiles[ind]
        if args.verbose > 0:
            print(ifile)
        imap = enmap.read_map(ifile)

        N = imap.shape[:-2]
        ndigits = [get_num_digits(n) for n in N]
        for i, map in enumerate(imap.preflat):
            I = np.unravel_index(i, N) if len(N) > 0 else []  # noqa
            comp = (
                "_" + "_".join(["%0*d" % (ndig, ind) for ndig, ind in zip(ndigits, I)])
                if len(N) > 0
                else ""
            )
            ofile = ifile[:-5] + args.suffix + comp + args.ext

            # Quantize it
            mask = map == args.mask
            qmap = pack(map, mask, nbyte=args.nbyte, quantum=args.quantum)
            img = Image.fromarray(qmap, mode="L")
            img.save(ofile)
